import { Server } from 'socket.io'
import { TSocket } from '../../utilities/Types'
import ZoneManager from '../../managers/ZoneManager'
import prisma from '../../utilities/Prisma'
import { AStar, type Node } from '../../utilities/Player/AStar'
import Rotation from '../../utilities/Player/Rotation'
import { ExtendedCharacter as Character } from '../../utilities/Types'
import { start } from 'node:repl'

interface SocketResponse {
  position_x: number
  position_y: number
}

// Map to store movement tokens for each character
const characterMoveTokens = new Map<number, Symbol>()

export default function setupCharacterMove(socket: TSocket, io: Server) {
  socket.on('character:move', async (data: SocketResponse) => {
    try {
      console.log('character:move requested', data)

      if (!socket.character) {
        console.error('character:move error', 'Character not found')
        return
      }

      // Cancel any ongoing movement for this character
      const moveToken = Symbol('moveToken')
      characterMoveTokens.set(socket.character.id, moveToken)

      const grid = await ZoneManager.getGrid(socket.character.zoneId)

      if (grid.length === 0) {
        console.error('character:move error', 'Grid not found')
        return
      }

      const start = { x: socket.character.position_x, y: socket.character.position_y }
      const end = { x: data.position_x, y: data.position_y }

      const path = AStar.findPath(start, end, grid)

      if (path.length > 0) {
        socket.character.isMoving = true
        io.in(socket.character.zoneId.toString()).emit('character:moved', socket.character)

        // Start movement in a non-blocking manner
        moveAlongPath(socket, io, path, grid, moveToken).catch(console.error)
      } else {
        console.log('character:move error', 'No valid path found')
      }
    } catch (error) {
      console.error('character:move error', error)
      if (socket.character) {
        socket.character.isMoving = false
        io.in(socket.character.zoneId.toString()).emit('character:moved', socket.character)
      }
    }
  })
}

async function moveAlongPath(socket: TSocket, io: Server, path: Node[], grid: number[][], moveToken: Symbol) {
  if (!socket.character) return

  const totalSteps = path.length
  const updateInterval = 50 // milliseconds between updates
  const totalDuration = totalSteps * 250 // total duration of movement

  for (let step = 0; step < totalSteps; step++) {
    const startTime = Date.now()

    while (Date.now() - startTime < 250) { // 250ms per tile
      if (characterMoveTokens.get(socket.character.id) !== moveToken) {
        console.log('Movement cancelled for character', socket.character.id)
        return
      }

      const progress = (Date.now() - startTime) / 250
      const currentPosition = interpolatePosition(path[step], path[step + 1] || path[step], progress)

      if (isObstacle(currentPosition, grid)) {
        console.log('Obstacle encountered at', currentPosition)
        break
      }

      const rotation = step < totalSteps - 1
        ? Rotation.calculate(path[step].x, path[step].y, path[step + 1].x, path[step + 1].y)
        : socket.character.rotation

      await updateCharacterPosition(socket.character, currentPosition.x, currentPosition.y, rotation)

      ZoneManager.updateCharacterInZone(socket.character.zoneId, socket.character)
      io.in(socket.character.zoneId.toString()).emit('character:moved', socket.character)

      await new Promise(resolve => setTimeout(resolve, updateInterval))
    }
  }

  if (socket.character) {
    socket.character.isMoving = false
    io.in(socket.character.zoneId.toString()).emit('character:moved', socket.character)
  }
}

function interpolatePosition(start: Node, end: Node, progress: number): Node {
  return {
    f: 0, g: 0, h: 0,
    x: start.x + (end.x - start.x) * progress,
    y: start.y + (end.y - start.y) * progress
  }
}

function isObstacle(position: Node, grid: number[][]): boolean {
  const x = Math.floor(position.x)
  const y = Math.floor(position.y)
  return grid[y] && grid[y][x] === 1
}

async function updateCharacterPosition(character: Character, x: number, y: number, rotation: number) {
  character.position_x = x
  character.position_y = y
  character.rotation = rotation

  await prisma.character.update({
    where: { id: character.id },
    data: { position_x: x, position_y: y, rotation }
  })
}